﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Build.Utilities;
using TestPlanWordWriter.Common.Word;
using Microsoft.Build.Framework;
using System.IO;
using Microsoft.TeamFoundation.Client;
using System.Threading;
using TestPlanWordWriter.Common;

namespace TestPlanWordWriter.Build
{
    public class RefreshDocumentTask : Task
    {
        [Required]
        public string ProjectName { get; set; }
        [Required]
        public string CollectionUri { get; set; }
        [Required]
        public ITaskItem InputDocumentFile { get; set; }
        [Required]
        public ITaskItem OutputDirectory { get; set; }
        [Required]
        public string OutputFileName { get; set; }
        [Required]
        public ITaskItem SettingsFile { get; set; }

        public override bool Execute()
        {
            bool validParameters = CheckParameters();
            if (validParameters)
            {
                DoRefresh();
            }
            else
            {
                Log.LogError("Invalid parameters");
            }

            return !Log.HasLoggedErrors;
        }

        private void LogFormat(string message, params object[] args)
        { 
            if (this.BuildEngine != null) 
            { 
                this.Log.LogMessage(message, args); 
            } 
            else 
            {
                Console.WriteLine(message, args); 
            } 
        }

        private void DoRefresh()
        {

            Uri projectCollectionUri = new Uri(this.CollectionUri);
            TfsTeamProjectCollection tpc = TfsTeamProjectCollectionFactory.GetTeamProjectCollection(projectCollectionUri);

            TestPlanDocWriterSettings settings = TestPlanDocWriterSettings.CreateFromFile(this.SettingsFile.ItemSpec);

            CancellationToken cancelToken = new CancellationToken();

            //Fill the content
            string outputFilePath = Path.Combine(this.OutputDirectory.ItemSpec, this.OutputFileName);
            using (Stream outputStream = new FileStream(outputFilePath, FileMode.Create, FileAccess.ReadWrite))
            {
                //Copy input file into output file before writing content
                using (Stream sourceStream = new FileStream(this.InputDocumentFile.ItemSpec, FileMode.Open, FileAccess.Read))
                {
                    StreamUtil.CopyStream(sourceStream, outputStream);
                }

                DocumentContentRefresh refresher = new DocumentContentRefresh();
                refresher.RefreshTestPlans(outputStream, tpc, this.ProjectName, settings, cancelToken);
            }
        }

        private bool CheckParameters()
        {
            if ((InputDocumentFile == null) || (!File.Exists(InputDocumentFile.ItemSpec)))
            {
                if (InputDocumentFile == null)
                {
                    Log.LogError("InputDocumentFile property must point to an existing file");
                }
                else
                {
                    Log.LogError("Input document file '{0}' does not exist", InputDocumentFile.ItemSpec);
                }
            }

            if ((SettingsFile == null) || (!File.Exists(SettingsFile.ItemSpec)))
            {
                if (SettingsFile == null)
                {
                    Log.LogError("SettingsFile property must point to an existing file");
                }
                else
                {
                    Log.LogError("SettingsFile '{0}' does not exist", SettingsFile.ItemSpec);
                }
            }

            if ((OutputDirectory == null) || (!Directory.Exists(OutputDirectory.ItemSpec)))
            {
                if (OutputDirectory == null)
                {
                    Log.LogError("OutputDirectory property must point to an existing directory");
                }
                else
                {
                    Log.LogError("Output directory '{0}' does not exist", OutputDirectory.ItemSpec);
                }
            }

            return !Log.HasLoggedErrors;
        }

    }
}
