﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Xml.Linq;
using System.IO;
using DocumentFormat.OpenXml.Packaging;
using DocumentFormat.OpenXml.Wordprocessing;
using DocumentFormat.OpenXml;
using Microsoft.TeamFoundation.TestManagement.Client;
using DocumentFormat.OpenXml.Drawing.Wordprocessing;

namespace TestPlanWordWriter.Common.Word
{
    public class WordPlanContentBuilder
    {
        private TestPlanDocWriterSettings _settings;
        private ContentControlIdentification _controlIdentification;
        private const int FIRST_COLUMN_WIDTH_IN_CM = 2;
        private string FAILED_COLOR = "FF0000";
        private string PASSED_COLOR = "00B050";
        private string WARNING_COLOR = "FF9900";

        public WordPlanContentBuilder(TestPlanDocWriterSettings settings, ContentControlIdentification controlIdentification)
        {
            _settings = settings;
            _controlIdentification = controlIdentification;
        }

        public void BuildPackageContent(TestPlanInfo planInfo, string stylesXml, Stream documentStream)
        {
            using (WordprocessingDocument processingDocument = WordprocessingDocument.Create(documentStream, DocumentFormat.OpenXml.WordprocessingDocumentType.Document))
            {
                // Add a new main document part.
                processingDocument.AddMainDocumentPart();

                // Create the Document DOM.             
                processingDocument.MainDocumentPart.Document = new DocumentFormat.OpenXml.Wordprocessing.Document(new Body());

                Body body = processingDocument.MainDocumentPart.Document.Body;
                StyleDefinitionsPart docStyleDefinitionsPart = processingDocument.MainDocumentPart.AddNewPart<StyleDefinitionsPart>("rId1");

                FillStyleDefinitionsPart(docStyleDefinitionsPart, stylesXml);

                FillCompositeElementContent(planInfo, body);

                processingDocument.Package.Flush();

                // Save changes to the main document part.             
                processingDocument.MainDocumentPart.Document.Save();
            }
        }

        private static void FillStyleDefinitionsPart(StyleDefinitionsPart docStyleDefinitionsPart, string stylesXml)
        {
            OpcStreamBuilder builder = new OpcStreamBuilder(stylesXml);
            using (MemoryStream openDocumentStream = new MemoryStream())
            {
                builder.BuildOpenDocumentStream(openDocumentStream);
                using (WordprocessingDocument sourceDocument = WordprocessingDocument.Open(openDocumentStream, true))
                {
                    docStyleDefinitionsPart.Styles = new Styles();
                    docStyleDefinitionsPart.Styles.Load(sourceDocument.MainDocumentPart.StyleDefinitionsPart);
                }
            }
        }

        public void FillCompositeElementContent(TestPlanInfo planInfo, OpenXmlCompositeElement mainContentContainer)
        {
            AppendTitleAndDescriptionParagraph(mainContentContainer, planInfo.Name, planInfo.Description, _settings.TestPlanStyleName, _settings.TestPlanDescriptionStyleName);

            FillSuiteChildrenContent(planInfo.RootSuite, mainContentContainer, 0);
        }

        private void FillSuiteContent(TestSuiteInfo suite, OpenXmlCompositeElement mainContentContainer, int suiteLevel)
        {
            string titleStyleName = null;
            string descriptionStyleName = null;
            if (suiteLevel <= _settings.TestSuiteLevelsSettings.Count)
            {
                titleStyleName = _settings.TestSuiteLevelsSettings[suiteLevel - 1].TestSuiteStyleName;
                descriptionStyleName = _settings.TestSuiteLevelsSettings[suiteLevel - 1].TestSuiteDescription;
            }

            AppendTitleAndDescriptionParagraph(mainContentContainer, suite.Title, suite.Description, titleStyleName, descriptionStyleName);

            FillSuiteChildrenContent(suite, mainContentContainer, suiteLevel);
        }

        private void FillSuiteChildrenContent(TestSuiteInfo suite, OpenXmlCompositeElement mainContentContainer, int suiteLevel)
        {
            var orderedList = suite.TestCases.OrderBy(x => x.Title);
            foreach (TestCaseInfo testCase in orderedList)
            {
                FillTestCaseContent(testCase, mainContentContainer);
            }

            //Recursive fill
            var orderedSuites = suite.InnerSuites.OrderBy(testSuite => testSuite.Title);
            foreach (TestSuiteInfo innerSuiteInfo in suite.InnerSuites)
            {
                int innerLevel = suiteLevel + 1;
                FillSuiteContent(innerSuiteInfo, mainContentContainer, innerLevel);
            }
        }

        private void AppendTitleAndDescriptionParagraph(OpenXmlCompositeElement mainContentContainer, string title, string description, string titleStyleName, string descriptionStyleName)
        {
            Paragraph suiteParagraph = new Paragraph();
            ParagraphProperties suiteParagraphProperties = new ParagraphProperties();

            ParagraphStyleId suiteParagraphStyleId = BuildStyleFromStyleName(titleStyleName);
            suiteParagraphProperties.AppendChild<ParagraphStyleId>(suiteParagraphStyleId);

            suiteParagraph.AppendChild<ParagraphProperties>(suiteParagraphProperties);
            suiteParagraph.AppendChild<Run>(new Run(new Text(title)));

            mainContentContainer.AppendChild<Paragraph>(suiteParagraph);

            if (!string.IsNullOrEmpty(description))
            {
                Paragraph suiteDescParagraph = new Paragraph();

                ParagraphProperties suiteDescParagraphProperties = new ParagraphProperties();
                ParagraphStyleId suiteDescParagraphStyleId = BuildStyleFromStyleName(descriptionStyleName);
                suiteDescParagraphProperties.AppendChild<ParagraphStyleId>(suiteDescParagraphStyleId);

                suiteDescParagraph.AppendChild<ParagraphProperties>(suiteDescParagraphProperties);
                suiteDescParagraph.AppendChild<Run>(new Run(new Text(description)));

                mainContentContainer.AppendChild<Paragraph>(suiteDescParagraph);
            }
        }

        private void FillTestCaseContent(TestCaseInfo testCase, OpenXmlCompositeElement mainContentContainer)
        {
            AppendTitleAndDescriptionParagraph(mainContentContainer, testCase.Title, testCase.Description, _settings.TestCaseStyleName, _settings.TestCaseDescriptionStyleName);

            if (_controlIdentification.ControlType == TestControlType.TestPlanControl)
            {
                if (testCase.TestSteps.Count > 0)
                {
                    FillTestCaseContent_Steps(testCase, mainContentContainer);
                }
            }
            else
            {
                //Has results in any testpoint
                if (testCase.TestPoints.Any(testPoint => testPoint.ResultsInfo.Count > 0))
                {
                    FillTestCaseContent_Results(testCase, mainContentContainer);
                }
            }
        }

        private void FillTestCaseContent_Results(TestCaseInfo testCase, OpenXmlCompositeElement mainContentContainer)
        {
            Table stepsTable = new Table();

            TableProperties stepsTableProperties = new TableProperties();
            TableStyle stepsTableStyle = BuildTableStyleFromStyleName(_settings.TestResultsTableStyleName);
            TableWidth stepsTableWidth = new TableWidth() { Width = "5000", Type = TableWidthUnitValues.Pct };

            TableGrid stepsTableGrid = new TableGrid();

            stepsTableProperties.Append(stepsTableStyle);
            stepsTableProperties.Append(stepsTableWidth);

            stepsTable.AppendChild<TableProperties>(stepsTableProperties);
            stepsTable.AppendChild<TableGrid>(stepsTableGrid);

            TableCell resultOutcomeCellHeader = new TableCell(new Paragraph(new Run(new Text(_settings.TableLabel_Results_ResultOutcome))));
            TableCell configurationNameCellHeader = new TableCell(new Paragraph(new Run(new Text(_settings.TableLabel_Results_ConfigurationName))));
            TableCell dateStartedCellHeader = new TableCell(new Paragraph(new Run(new Text(_settings.TableLabel_Results_DateStarted))));
            TableCell dateCompletedCellHeader = new TableCell(new Paragraph(new Run(new Text(_settings.TableLabel_Results_DateCompleted))));
            TableCell commentCellHeader = new TableCell(new Paragraph(new Run(new Text(_settings.TableLabel_Results_Comment))));
            TableRow headerRow = new TableRow(resultOutcomeCellHeader, configurationNameCellHeader, dateStartedCellHeader, dateCompletedCellHeader, commentCellHeader);
            stepsTable.AppendChild<TableRow>(headerRow);

            foreach (TestPointInfo pointInfo in testCase.TestPoints)
            {
                var orderedResultsByDate = pointInfo.ResultsInfo.OrderByDescending(x => x.DateStarted);
                foreach (TestResultInfo resultInfo in orderedResultsByDate)
                {
                    //TODO: Add Icon image column using result outcome
                    Color resultOutcomeColor = new Color();
                    if (resultInfo.ResultOutcome == TestOutcome.Passed)
                    {
                        resultOutcomeColor.Val = PASSED_COLOR;
                    }
                    else if (resultInfo.ResultOutcome == TestOutcome.Failed)
                    {
                        resultOutcomeColor.Val = FAILED_COLOR;
                    }
                    else
                    {
                        resultOutcomeColor.Val = WARNING_COLOR;
                    }

                    TableCell resultOutcomeCell = new TableCell(new Paragraph(
                        new Run(
                            new RunProperties(resultOutcomeColor), 
                            new Text(resultInfo.ResultOutcome.ToString()))));
                    TableCell configurationNameCell = new TableCell(new Paragraph(new Run(new Text(pointInfo.ConfigurationName))));
                    TableCell dateStartedCell = new TableCell(new Paragraph(new Run(new Text(resultInfo.DateStarted.ToString()))));
                    TableCell dateCompletedCell = new TableCell(new Paragraph(new Run(new Text(resultInfo.DateCompleted.ToString()))));
                    TableCell commentCell = new TableCell(new Paragraph(new Run(new Text(resultInfo.Comment))));

                    TableRow stepRow = new TableRow(resultOutcomeCell, configurationNameCell, dateStartedCell, dateCompletedCell, commentCell);
                    stepsTable.AppendChild<TableRow>(stepRow);
                }
            }

            mainContentContainer.AppendChild<Table>(stepsTable);
        }

        private void FillTestCaseContent_Steps(TestCaseInfo testCase, OpenXmlCompositeElement mainContentContainer)
        {
            Table stepsTable = new Table();

            TableProperties stepsTableProperties = new TableProperties();
            TableStyle stepsTableStyle = BuildTableStyleFromStyleName(_settings.TestStepsTableStyleName);
            TableWidth stepsTableWidth = new TableWidth() { Width = "5000", Type = TableWidthUnitValues.Pct };

            ////2cm
            float firstColumnWidth = FIRST_COLUMN_WIDTH_IN_CM * 28.3157f;

            TableGrid stepsTableGrid = new TableGrid();
            GridColumn gridColumn1 = new GridColumn() { Width = Convert.ToInt32(firstColumnWidth).ToString() };
            GridColumn gridColumn2 = new GridColumn();
            GridColumn gridColumn3 = new GridColumn();

            stepsTableProperties.Append(stepsTableStyle);
            stepsTableProperties.Append(stepsTableWidth);

            stepsTable.AppendChild<TableProperties>(stepsTableProperties);
            stepsTable.AppendChild<TableGrid>(stepsTableGrid);

            TableCell indexCellHeader = new TableCell(new Paragraph(new Run(new Text(_settings.TableLabel_Step))));
            TableCell titleCellHeader = new TableCell(new Paragraph(new Run(new Text(_settings.TableLabel_Action))));
            TableCell expectedResultCellHeader = new TableCell(new Paragraph(new Run(new Text(_settings.TableLabel_ExpectedResult))));
            TableRow headerRow = new TableRow(indexCellHeader, titleCellHeader, expectedResultCellHeader);
            stepsTable.AppendChild<TableRow>(headerRow);

            int i = 1;
            foreach (TestStepInfo stepInfo in testCase.TestSteps)
            {
                TableCell indexCell = new TableCell(new Paragraph(new Run(new Text(i.ToString()))));
                TableCell titleCell = new TableCell(new Paragraph(new Run(new Text(stepInfo.Title))));
                TableCell expectedResultCell = new TableCell(new Paragraph(new Run(new Text(stepInfo.ExpectedResult))));
                i++;
                TableRow stepRow = new TableRow(indexCell, titleCell, expectedResultCell);
                stepsTable.AppendChild<TableRow>(stepRow);
            }

            mainContentContainer.AppendChild<Table>(stepsTable);
        }

        private TableStyle BuildTableStyleFromStyleName(string styleName)
        {
            TableStyle paragraphStyleId = new TableStyle()
            {
                Val = StyleNameFormatter.ConvertVSTONameToOpenName(styleName)
            };
            return paragraphStyleId;
        }

        private ParagraphStyleId BuildStyleFromStyleName(string styleName)
        {
            ParagraphStyleId paragraphStyleId = new ParagraphStyleId() 
            { 
                Val = StyleNameFormatter.ConvertVSTONameToOpenName(styleName) 
            };
            return paragraphStyleId;
        }
    }
}
