﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.TeamFoundation.Client;
using Microsoft.TeamFoundation.TestManagement.Client;
using System.Diagnostics;
using Microsoft.TeamFoundation.WorkItemTracking.Client;
using System.Threading;

namespace TestPlanWordWriter.Common
{
    public class TestPlanInfoSimpleBuilder
    {
        public delegate void TestSuitesInfoProgressChangedEventHandler(object sender, TestSuitesInfoProgressChangedEventArgs args);
        public delegate void TestCasesInfoProgressChangedEventHandler(object sender, TestCasesInfoProgressChangedEventArgs args);
        public delegate void BeforeTestPlanProcessEventHandler(object sender, BeforeTestPlanProcessEventArgs args);
        public delegate void BeforeTestSuiteProcessEventHandler(object sender, BeforeTestSuiteProcessEventArgs args);
        public delegate void BeforeTestCaseProcessEventHandler(object sender, BeforeTestCaseProcessEventArgs args);

        public event TestSuitesInfoProgressChangedEventHandler TestSuitesProgressChanged;
        public event TestCasesInfoProgressChangedEventHandler TestCasesProgressChanged;
        public event BeforeTestPlanProcessEventHandler BeforeTestPlanProcess;
        public event BeforeTestSuiteProcessEventHandler BeforeTestSuiteProcess;
        public event BeforeTestCaseProcessEventHandler BeforeTestCaseProcess;

        public int TotalSuites { get; private set; }
        public int ProcessedSuites { get; private set; }
        public int TotalTestCases { get; private set; }
        public int ProcessedTestCases { get; private set; }

        public TestPlanInfo BuildPlanInfo(TfsTeamProjectCollection tpc, string teamProjectName, int testPlanId, CancellationToken cancelToken)
        {
            ITestManagementService test_service = (ITestManagementService)tpc.GetService(typeof(ITestManagementService));
            ITestManagementTeamProject testProject = test_service.GetTeamProject(teamProjectName);

            this.ProcessedSuites = 0;
            this.TotalSuites = testProject.TestSuites.Query("SELECT * FROM TestSuite WHERE PlanId = " + testPlanId.ToString()).Count;

            ITestPlan testPlan = testProject.TestPlans.Find(testPlanId);

            OnBeforeTestPlanProcess(testPlan);

            TestSuiteInfo rootSuiteInfo = BuildTestSuiteInfoRecursive(tpc, testPlan.RootSuite, cancelToken);

            TestPlanInfo planInfo = new TestPlanInfo(testPlan, rootSuiteInfo);
            return planInfo;
        }

        private TestSuiteInfo BuildTestSuiteInfoRecursive(TfsTeamProjectCollection tpc, ITestSuiteBase testSuite, CancellationToken cancelToken)
        {
            OnBeforeTestSuiteProcess(testSuite);

            TestSuiteInfo info = new TestSuiteInfo();

            if (testSuite.TestSuiteType == TestSuiteType.RequirementTestSuite)
            {
                IRequirementTestSuite requirementSuite = testSuite as IRequirementTestSuite;

                int workItemId = requirementSuite.RequirementId;
                WorkItem userStory = TfsGateway.GetWorkItem(tpc, workItemId);

                info.Id = userStory.Id;
                info.Description = userStory.Description;
                info.Title = userStory.Title;
            }
            else
            {
                info.Id = testSuite.Id;
                info.Description = testSuite.Description;
                info.Title = testSuite.Title;
            }

            if (testSuite.TestSuiteType == TestSuiteType.StaticTestSuite)
            {
                IStaticTestSuite staticSuite = testSuite as IStaticTestSuite;
                foreach (ITestSuiteBase subSuite in staticSuite.SubSuites)
                {
                    TestSuiteInfo childSuiteinfo = BuildTestSuiteInfoRecursive(tpc, subSuite, cancelToken);
                    info.InnerSuites.Add(childSuiteinfo);
                }
            }

            info.TestCases = BuildTestCases(testSuite, cancelToken);

            OnTestSuitesProgressChanged();

            return info;
        }

        public List<TestCaseInfo> BuildTestCases(ITestSuiteBase testSuite, CancellationToken cancelToken)
        {
            ITestSuiteEntryCollection testCases = testSuite.TestCases;

            this.ProcessedTestCases = 0;
            this.TotalTestCases = testCases.Count;

            List<TestCaseInfo> innerTestCases = new List<TestCaseInfo>();
            foreach (ITestSuiteEntry testSuiteEntry in testCases)
            {
                cancelToken.ThrowIfCancellationRequested();

                TestCaseInfo info = BuildTestCaseInfo(testSuiteEntry.TestCase);
                innerTestCases.Add(info);

                OnTestCasesProgressChanged();
            }
            return innerTestCases;
        }

        private TestCaseInfo BuildTestCaseInfo(ITestCase testCase)
        {
            OnBeforeTestCaseProcess(testCase);

            List<TestStepInfo> testSteps = new List<TestStepInfo>();

            TestActionCollection actions = testCase.Actions;
            foreach (ITestAction action in actions)
            {
                FillTestStepsFromAction(testSteps, action);
            }

            TestCaseInfo info = new TestCaseInfo(testCase, testSteps);
            return info;
        }

        private void FillTestStepsFromAction(List<TestStepInfo> testSteps, ITestAction action)
        {
            if (action is ITestStep)
            {
                ITestStep step = action as ITestStep;
                TestStepInfo info = new TestStepInfo(step);
                testSteps.Add(info);
            }
            else if (action is ISharedStepReference)
            {
                //Recursive search
                ISharedStepReference sharedStepReference = action as ISharedStepReference;
                ISharedStep sharedStep = sharedStepReference.FindSharedStep();
                foreach (ITestAction sharedStepAction in sharedStep.Actions)
                {
                    FillTestStepsFromAction(testSteps, sharedStepAction);
                }
            }
        }

        private void OnTestSuitesProgressChanged()
        {
            this.ProcessedSuites++;
            if (TestSuitesProgressChanged != null)
            {
                TestSuitesInfoProgressChangedEventArgs args = new TestSuitesInfoProgressChangedEventArgs(this.ProcessedSuites, this.TotalSuites);
                TestSuitesProgressChanged(this, args);
            }
        }

        private void OnTestCasesProgressChanged()
        {
            this.ProcessedTestCases++;
            if (TestCasesProgressChanged != null)
            {
                TestCasesInfoProgressChangedEventArgs args = new TestCasesInfoProgressChangedEventArgs(this.ProcessedTestCases, this.TotalTestCases);
                TestCasesProgressChanged(this, args);
            }
        }

        private void OnBeforeTestPlanProcess(ITestPlan testPlan)
        {
            if (BeforeTestPlanProcess != null)
            {
                BeforeTestPlanProcessEventArgs args = new BeforeTestPlanProcessEventArgs(testPlan);
                BeforeTestPlanProcess(this, args);
            }
        }

        private void OnBeforeTestSuiteProcess(ITestSuiteBase testSuite)
        {
            if (BeforeTestSuiteProcess != null)
            {
                BeforeTestSuiteProcessEventArgs args = new BeforeTestSuiteProcessEventArgs(testSuite);
                BeforeTestSuiteProcess(this, args);
            }
        }

        private void OnBeforeTestCaseProcess(ITestCase testCase)
        {
            if (BeforeTestCaseProcess != null)
            {
                BeforeTestCaseProcessEventArgs args = new BeforeTestCaseProcessEventArgs(testCase);
                BeforeTestCaseProcess(this, args);
            }
        }
    }
}
